/*
 * Sfontview - A simple lightweight TrueType font viewer.
 * Copyright (C) 2012  Thunor <thunorsif@hotmail.com>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <SDL/SDL.h>
#include <SDL/SDL_main.h>
#include <SDL/SDL_ttf.h>

#define TRUE 1
#define FALSE 0
#define TARGET "sfontview"
#define TITLE "Sfontview"
#define VERSION "0.1.1"
#define TTF_RENDER TTF_RenderUTF8_Shaded
/* User configurable values */
#define I18N_UNKNOWN "Unknown"
#define I18N_PANGRAM "The quick brown fox jumps over the lazy dog"
#define BG_COLOUR 0xffffff
#define FG_COLOUR 0x101010
#define SCREEN_WIDTH 360
#define SCREEN_HEIGHT 340
/* Index 0 is family/style, 1 is abc/ABC/012 and 2 to 7 are pangrams */
#define FONT_SIZES 12, 20, 8, 12, 16, 24, 48, 64

/* Local variables */
SDL_Surface *img_main = NULL;
SDL_Surface *img_text = NULL;
SDL_Surface *img_icon = NULL;
TTF_Font *fonts[8];

/* Function prototypes */
char *execute_command(char *command);
void clean_up_before_exit(void);

/* README
 * ======
 * 
 * Compile with:
 * gcc -Wall sfontview.c -o sfontview `sdl-config --cflags --libs` -lSDL_ttf
 * 
 * Strip with:
 * strip --strip-unneeded sfontview
 * 
 * Place the sfontview binary into ~/my-applications/bin or /usr/local/bin.
 * Place sfontview.bmp into /usr/share/pixmaps or /usr/local/share/pixmaps.
 * 
 * The font family and style are rendered using the user's default font
 * which can be found using fc-match and this should be available since
 * it's part of the fontconfig package, but if for some reason it's not
 * then the text will be rendered using the user supplied font.
 * 
 * Don't forget that ESC will also close the window so you can quickly
 * click through lots of fonts with one hand hovering over the ESC key.
 * 
 * ChangeLog
 * =========
 * 
 * Wed Oct 31 2012 Thunor <thunorsif@hotmail.com>
 * 0.1.1
 * Added an optional window manager icon to be placed into
 * /usr/share/pixmaps or /usr/local/share/pixmaps.
 * Included the SDL_main stuff.
 * 
 * Sat Oct 27 2012 Thunor <thunorsif@hotmail.com>
 * 0.1.0
 * Initial release.
 */

#ifdef __cplusplus
extern "C"
#endif
int main(int argc, char *argv[])
{
	SDL_Color bgcolor, fgcolor;
	SDL_Event event;
	SDL_Rect destrect;
	SDL_Surface *screen;
	char command[256];
	char *font_family;
	char *font_filename;
	char font_filename_default[256];
	char font_filename_user[256];
	char *font_style;
	char *retval;
	char window_title[256];
	int count;
	int font_sizes[] = {FONT_SIZES};
	int quit = FALSE;

	/* Initialise variables */
	for (count = 0; count < 8; count++)
		fonts[count] = NULL;

	/* Process command-line arguments */
	if (argc < 2 || (argc > 1 &&
		((!strcmp(argv[1], "--help")) || !strcmp(argv[1], "-h")))) {
		fprintf(stdout, TITLE " " VERSION " (C) 2012 Thunor\n");
		fprintf(stdout, "Usage: " TARGET " [FILE]\n\n");
		fprintf(stdout, "Where FILE is a TrueType font file.\n\n");
		return 1;
	}

	/* Get font file from command-line args */
	strcpy(font_filename_user, argv[1]);

	/* Locate the user's default font */
	strcpy(font_filename_default,
		execute_command("DFONT=`fc-match`; echo ${DFONT%%:*}"));
	for (count = 0; count < 3; count++) {
		if (count == 0) {
			strcpy(command, "find /usr/share/fonts -name \"");
		} else if (count == 1) {
			strcpy(command, "find /usr/X11R6/lib/X11/fonts/ -name \"");
		} else {
			strcpy(command, "find ~/.fonts -name \"");
		}
		strcat(command, font_filename_default);
		strcat(command, "\" | head -1");
		retval = execute_command(command);
		if ((strcmp(retval, "")) != 0) break;
		if (count == 2) {
			fprintf(stderr, "Couldn't locate default font %s\n",
				font_filename_default);
			retval = font_filename_user;
		}
	}
	strcpy(font_filename_default, retval);

	/* Initialise SDL */
	if (SDL_Init(SDL_INIT_VIDEO)) {
		fprintf(stderr, "Couldn't initialise SDL: %s", SDL_GetError());
		return 1;
	}

	atexit(clean_up_before_exit);

	/* Set the window manager icon if present */
	img_icon = SDL_LoadBMP("/usr/share/pixmaps/" TARGET ".bmp");
	if (img_icon == NULL)
		img_icon = SDL_LoadBMP("/usr/local/share/pixmaps/" TARGET ".bmp");
	if (img_icon) {
		/* Set the transparent colour */
		if (SDL_SetColorKey(img_icon, SDL_SRCCOLORKEY,
			SDL_MapRGB(img_icon->format, 0xff, 0x00, 0xff)) < 0) {
			fprintf(stderr, "Couldn't set surface colour key: %s\n",
				SDL_GetError());
		} else {
			SDL_WM_SetIcon(img_icon, NULL);
		}
	}

	/* Set the window manager caption */
	count = strlen(font_filename_user) - 1;
	while (count) {
		if (*(font_filename_user + count - 1) == '/') break;
		count--;
	}
	strcpy(window_title, TITLE " - ");
	strcat(window_title, font_filename_user + count);
	SDL_WM_SetCaption(window_title, TITLE);

	/* Set the video mode */
	if ((screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, 0,
		SDL_SWSURFACE)) == NULL) {
		fprintf(stderr, "Couldn't set video mode: %s", SDL_GetError());
		exit(1);
	}

	/* SDL_ttf initialisation */
	if (TTF_Init()) {
		fprintf(stderr, "Couldn't initialise SDL_ttf: %s\n", TTF_GetError());
		exit(1);
	}

	/* Load the fonts */
	font_filename = font_filename_default;
	for (count = 0; count < 8; count++) {
		fonts[count] = TTF_OpenFont(font_filename, font_sizes[count]);
		if (fonts[count] == NULL) {
			fprintf(stderr, "Couldn't load font %s: %s\n", font_filename,
				TTF_GetError());
			exit(1);
		}
		font_filename = font_filename_user;
	}

	/* Get font information (could return NULL) */
	if ((font_family = TTF_FontFaceFamilyName(fonts[1])) == NULL)
		font_family = I18N_UNKNOWN;
	if ((font_style = TTF_FontFaceStyleName(fonts[1])) == NULL)
		font_style = I18N_UNKNOWN;

	/* Initialise some colours */
	bgcolor.r = BG_COLOUR >> 16 & 0xff;
	bgcolor.g = BG_COLOUR >> 8 & 0xff;
	bgcolor.b = BG_COLOUR & 0xff;
	fgcolor.r = FG_COLOUR >> 16 & 0xff;
	fgcolor.g = FG_COLOUR >> 8 & 0xff;
	fgcolor.b = FG_COLOUR & 0xff;

	/* Build the main image */
	if ((img_main = SDL_CreateRGBSurface(screen->flags,
		SCREEN_WIDTH - 10, SCREEN_HEIGHT - 10,
		screen->format->BitsPerPixel,
		screen->format->Rmask, screen->format->Gmask,
		screen->format->Bmask, screen->format->Amask)) == NULL) {
		fprintf(stderr, "Couldn't create RGB surface: %s", SDL_GetError());
		exit(1);
	}
	/* Wipe the surface */
	SDL_FillRect(img_main, NULL, SDL_MapRGB(img_main->format, bgcolor.r,
		bgcolor.g, bgcolor.b));
	/* Write the text */
	destrect.x = destrect.y = destrect.w = destrect.h = 0;
	for (count = 0; count < 13; count++) {
		if (count == 0) {
			/* Write font family */
			img_text = TTF_RENDER(fonts[0], font_family,
				fgcolor, bgcolor);
		} else if (count == 1) {
			/* Write font style */
			img_text = TTF_RENDER(fonts[0], font_style,
				fgcolor, bgcolor);
		} else if (count == 3) {
			/* Write lowercase a-z */
			img_text = TTF_RENDER(fonts[1],
				"abcdefghijklmnopqrstuvwxyz", fgcolor, bgcolor);
		} else if (count == 4) {
			/* Write uppercase A-Z */
			img_text = TTF_RENDER(fonts[1],
				"ABCDEFGHIJKLMNOPQRSTUVWXYZ", fgcolor, bgcolor);
		} else if (count == 5) {
			/* Write numbers etc. */
			img_text = TTF_RENDER(fonts[1],
				"0123456789.:,;(*!?'/\\\")£$€%^&-+@", fgcolor, bgcolor);
		} else if (count >= 7) {
			/* Write pangrams */
			img_text = TTF_RENDER(fonts[count - 5],
				I18N_PANGRAM, fgcolor, bgcolor);
		}
		if (count == 2 || count == 6) {
			/* Draw a horizontal line */
			destrect.y = destrect.y + destrect.h;
			destrect.y = destrect.y + 2;
			destrect.w = img_main->w; destrect.h = 1;
			SDL_FillRect(img_main, &destrect, SDL_MapRGB(img_main->format,
				fgcolor.r, fgcolor.g, fgcolor.b));
			destrect.y = destrect.y + 2;
		} else {
			/* Blit */
			destrect.y = destrect.y + destrect.h;
			destrect.w = img_text->w; destrect.h = img_text->h;
			SDL_BlitSurface(img_text, NULL, img_main, &destrect);
			SDL_FreeSurface(img_text); img_text = NULL;
		}
	}

	/* Main loop */
	while (!quit) {

		while (SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_ESCAPE)
						quit = TRUE;
					break;
				case SDL_QUIT:
					quit = TRUE;
					break;
			}
		}

		/* Wipe the screen */
		SDL_FillRect(screen, NULL, SDL_MapRGB(screen->format,
			bgcolor.r, bgcolor.g, bgcolor.b));

		/* Blit the main image to the screen */
		destrect.x = destrect.y = 5;
		destrect.w = img_main->w; destrect.h = img_main->h;
		SDL_BlitSurface(img_main, NULL, screen, &destrect);

		/* Update the screen */
		SDL_Flip(screen);

		/* ~20fps is more than enough */
		SDL_Delay(50);
	}

	/* atexit() will call clean_up_before_exit() on return */
	return 0;
}

/**********************************************************************
 * Execute Command                                                    *
 **********************************************************************/
/* Get one line of output from a shell command.
 * 
 * On entry: command = the command to execute
 *  On exit: returns a pointer to valid output or a null string */

char *execute_command(char *command)
{
	FILE *fp;
	static char line[512];
	int count;

	strcpy(line, "");

	if ((fp = popen(command, "r"))) {
		/* Just one line */
		if ((fgets(line, 512, fp))) {
			/* Enforce end of string in case of max chars read */
			line[512 - 1] = 0;
			/* Remove the trailing [CR]LFs */
			for (count = strlen(line) - 1; count >= 0; count--)
				if (line[count] == 13 || line[count] == 10)
					line[count] = 0;
		}
		/* Close the file */
		pclose(fp);
	} else {
		fprintf(stderr, "%s(): Couldn't execute command %s\n",
			__func__, command);
	}

	return line;
}

/**********************************************************************
 * Clean Up Before Exit                                               *
 **********************************************************************/
/* This is registered with atexit() once SDL is initialised */

void clean_up_before_exit(void)
{
	int count;

	/* Free SDL surfaces if they are not NULL */
	if (img_main) SDL_FreeSurface(img_main);
	if (img_text) SDL_FreeSurface(img_text);
	if (img_icon) SDL_FreeSurface(img_icon);

	/* Free the fonts if not NULL */
	for (count = 0; count < 8; count++)
		if (fonts[count]) TTF_CloseFont(fonts[count]);

	TTF_Quit();
	SDL_Quit();
}
